import math
import subprocess
import threading
import random

from . import book

from collections import namedtuple

Evaluation = namedtuple("Evaluation", "score wdl depth nodes multipv pv")

Score_CentiPawn = namedtuple("CentiPawn", "value")
Score_Mate = namedtuple("Mate", "moves")
Score_CentiPawn.invert = lambda score: Score_CentiPawn(-score.value)
Score_Mate.invert = lambda score: Score_Mate(-score.moves)
Score_CentiPawn.__sub__ = lambda s1, s2: s1.value-s2.value
Score_Mate.__sub__ = lambda m1, m2: m1.moves-m2.moves

Request_Analyze = namedtuple("Analyze", "position white limit callback callback_args")
Request_Quit = namedtuple("Quit", "")

BestMove = namedtuple("BestMove", "uci")

def score_eval(score, depth=None):
    if isinstance(score, Score_CentiPawn):
        return f"{score.value/100:.2f}/{depth}" if depth is not None else f"{score.value/100:.2f}"
    elif isinstance(score, Score_Mate):
        return f"#{score.moves}/{depth}" if depth is not None else f"{score.moves}"
    else:
        raise ValueError("Not a valid score: {score}")

def score_winpercent(score):
    if isinstance(score, Score_CentiPawn):
        return 1/(1 + 10**(-score.value/400))
    elif isinstance(score, Score_Mate):
        return math.copysign(1.0, score.moves)
    else:
        raise ValueError(f"Not a valid score: {score}")

def uci_value_fmt(value):
    return ("true" if value else "false") if isinstance(value, bool) else str(value)

def score_of_string(score):
    if score.startswith("cp "):
        return Score_CentiPawn(int(score[3:]))
    if score.startswith("mate "):
        return Score_Mate(int(score[5:]))
    raise ValueError("Bad score format: " + repr(score))

class Engine:
    def __init__(self, exepath, options={}, maxdepth=None, maxtime=None, maxnodes=None, evaldb=None):
        self.maxdepth = maxdepth
        self.maxtime = maxtime
        self.maxnodes = maxnodes
        self.exepath = exepath
        self.ucioptions = options
        self.evaldb = evaldb
        self.quit_requested = False
        self.request = None
        self.sent_requests = []
        self.writer_cond = threading.Condition()
        self.writer_thread = threading.Thread(target=self._writer, name="an-writer", daemon=True)
        self.reader_thread = threading.Thread(target=self._reader, name="an-reader", daemon=True)
        self.id = {}
        self.process = subprocess.Popen(
                exepath,
                encoding="utf-8",
                universal_newlines=True,
                bufsize=1,
                stdin=subprocess.PIPE,
                stdout=subprocess.PIPE,
                stderr=subprocess.DEVNULL)
        self.writer_thread.start()
        self.reader_thread.start()

    def quit(self):
        with self.writer_cond:
            self.request = Request_Quit()
            self.writer_cond.notify()
        self.process.kill()
        self.writer_thread.join()
        self.reader_thread.join()

    def __del__(self):
        self.quit()

    def analyze(self, board, callback, *args):
        if self.evaldb:
            evals = self.evaldb(board)
            if evals:
                bestmove = None
                for depth, score, pv in evals:
                    if pv:
                        pv = pv.split()
                        bestmove = pv[0]
                    else:
                        pv = []
                    callback(Evaluation(score_of_string(score), None, depth, None, 1, pv), *args)
                assert(bestmove is not None)
                callback(BestMove(bestmove), *args)
                return
        position = "startpos moves " + " ".join(move.uci() for move in board.move_stack)
        limit = f"nodes {self.maxnodes}" if self.maxnodes is not None else \
                f"movetime {int(self.maxtime*1000)}" if self.maxtime is not None else \
                f"depth {self.maxdepth}" if self.maxdepth is not None else \
                f"infinite"
        with self.writer_cond:
            self.request = Request_Analyze(position, board.turn, limit, callback, args)
            self.writer_cond.notify()

    def play(self, board, callback, *args):
        self.analyze(board, self._play_callback, callback, *args)

    def _play_callback(self, result, callback, *args):
        if isinstance(result, BestMove):
            callback(result.uci, *args)

    def _writer(self):
        f = self.process.stdin
        f.write("uci\n")
        for k, v in self.ucioptions.items():
            f.write(f"setoption name {k} value {uci_value_fmt(v)}\n")
        f.write("isready\n")
        while True:
            with self.writer_cond:
                while self.request is None:
                    self.writer_cond.wait()
                req = self.request
                self.request = None
                self.sent_requests.append(req)
            if isinstance(req, Request_Quit):
                f.close()
                #f.write("stop\nisready\nquit\n")
                break
            if isinstance(req, Request_Analyze):
                f.write(f"stop\nposition {req.position}\ngo {req.limit}\n")

    def _reader(self):
        for line in self.process.stdout:
            if isinstance(self.request, Request_Quit):
                break
            words = line.split()
            if len(words) == 0:
                continue
            elif words[0] == "info":
                try:
                    i = words.index("score") + 1
                    if words[i] == "cp":
                        score = Score_CentiPawn(int(words[i + 1]))
                    elif words[i] == "mate":
                        score = Score_Mate(int(words[i + 1]))
                    else:
                        raise ValueError
                    depth = int(words[words.index("depth") + 1])
                    nodes = int(words[words.index("nodes") + 1])
                    pv = words[words.index("pv") + 1:]
                    try:
                        multipv = int(words[words.index("multipv") + 1])
                    except ValueError:
                        multipv = 1
                    req = self.sent_requests[0]
                    try:
                        i = words.index("wdl")
                        if req.white:
                            wdl = (int(words[i + 1]), int(words[i + 2]), int(words[i + 3]))
                        else:
                            wdl = (int(words[i + 3]), int(words[i + 2]), int(words[i + 1]))
                    except ValueError:
                        wdl = None
                    req.callback(Evaluation(
                            score if req.white else score.invert(),
                            wdl, depth, nodes, multipv, pv),
                        *req.callback_args)
                except ValueError as e:
                    # No "score", "depth", or "nodes".
                    pass
            elif words[0] == "bestmove":
                req = self.sent_requests[0]
                req.callback(BestMove(words[1]), *req.callback_args)
                # Mutation of self.sent_requests is not guarded
                # by a lock here because Python threads are not
                # real.
                self.sent_requests.pop(0)
            elif words[0] == "readyok":
                pass
            elif words[0] == "id":
                self.id[words[1]] = " ".join(words[2:])
            if isinstance(self.request, Request_Quit):
                break

class LossyEngine:
    def __init__(self, engine, maxloss=90):
        self.engine = engine
        self.maxloss = maxloss
        self.multipv = {}

    def play(self, board, callback, *args):
        self.engine.analyze(board, self._engine_callback, board.fen(), board.turn, callback, *args)

    def quit(self):
        self.engine.quit()

    def _engine_callback(self, x, fen, turn, callback, *args):
        if isinstance(x, BestMove):
            #print(f"LossyEngine: choosing from {self.multipv.values()}")
            moves = [move
                for score, move in self.multipv.values()
                if isinstance(score, Score_Mate) and score.moves >= 0]
            if moves:
                move = random.choice(moves)
                #print(f"LossyEngine: {moves} => {move}")
                self.multipv.clear()
                callback(move, *args)
                return
            moves = [(score.value, move)
                for score, move in self.multipv.values()
                if isinstance(score, Score_CentiPawn)]
            if moves:
                bestv = max(scorev for scorev, move in moves)
                moves = [move
                    for scorev, move in moves
                    if scorev >= bestv - self.maxloss]
                move = random.choice(moves)
                #print(f"LossyEngine: {moves} => {move}")
                self.multipv.clear()
                callback(move, *args)
                return
            #print(f"LossyEngine: => {x.uci}")
            self.multipv.clear()
            callback(x.uci, *args)
        elif isinstance(x, Evaluation):
            #print(f"Eval[{fen}, {turn}] = {x}")
            move = x.pv[0]
            self.multipv[x.multipv] = (x.score if turn else x.score.invert(), move)

def of_spec(spec):
    eng = Engine(exepath=spec["bin"],
            options=spec["options"],
            maxnodes=spec["limit"].get("maxnodes", None),
            maxdepth=spec["limit"].get("maxdepth", None),
            maxtime=spec["limit"].get("maxtime", None),
            evaldb=book.evaldb() if spec.get("eval_book", False) else None)
    if "maxloss" in spec:
        eng = LossyEngine(eng, spec["maxloss"])
    return eng



#!/usr/bin/env python3

import os
os.environ.setdefault("ESCDELAY", "25")

import chess
import chess.pgn
import curses
import datetime
import json
import os.path
import random
import re
import sys
import time

# All of the following should be relative imports, but in Python 3
# the "main module" if forbidden from using them, so if one
# wants to import this file (as we do in the appimage) instead
# of running it directly, PYTHONPATH must be set so that all of
# these files are in it, and come first.

from . import engine
from . import imui
from . import ecodb
from . import book

### Configuration

progdir = os.path.dirname(__file__)
config_kwargs = {
    "bin": os.path.join(progdir, "data"),
    "data": os.path.join(progdir, "data"),
    "date": datetime.date.today().strftime("%Y-%m-%d")
}

def config_subs(value):
    """Recursively substitute {variables} in a config dict."""
    if isinstance(value, list):
        return [config_subs(x) for x in value]
    elif isinstance(value, dict):
        return {k:config_subs(v) for k,v in value.items()}
    elif isinstance(value, str):
        return value.format_map(config_kwargs)
    elif isinstance(value, int):
        return value
    else:
        raise ValueError(f"Can't substitute variables in {value!r}")

def dict_overlay(dst, src):
    """
    Recursively apply the source dict on top of the destination
    dict: merge dictionaries, overwrite non-dictionary values.
    """
    for key, value in src.items():
        if isinstance(value, dict):
            oldvalue = dst.get(key, None)
            if isinstance(oldvalue, dict):
                dict_overlay(oldvalue, value)
            else:
                dst[key] = value
        else:
            dst[key] = value

def load_config(*filenames):
    """
    Load configuration dict from several files, return overlayed
    result.
    """
    conf = {}
    for fn in filenames:
        try:
            with open(fn, "r") as f:
                o = json.load(f)
            dict_overlay(conf, o)
        except OSError:
            pass
    return conf

def default_config():
    return config_subs(load_config(
        config_subs("{data}/bchess.conf"),
        os.path.expanduser("~/.config/bchess.conf"),
        os.path.expanduser("~/.bchess.conf")))

### MISC UI

def FormatText(text, attr=0):
    maxw = im.text_width
    for i, para in enumerate(text.split("\n")):
        if i > 0:
            im.VSpace(1)
        line = []
        linew = 0
        for word in para.split():
            if linew + 1 + len(word) > maxw:
                im.Text(" ".join(line), attr=attr)
                line = [word]
                linew = len(word)
            else:
                line.append(word)
                linew += 1 + len(word)
        if line:
            im.Text(" ".join(line), attr=attr)

UI_RENDERER = None
def set_scene(renderer):
    assert callable(renderer)
    global UI_RENDERER
    UI_RENDERER = renderer
    im.want_refresh = True

### CHESS UI

rx_move = re.compile("^([QKNRBqknrb])?(?:([a-h])([1-8])?)?([x:])?(?:([a-h])([1-8])?)?=?([QNRBqnrb])?[#+]?$")
def highlight_san_move(text, board=None):
    """
    Parse a move in SAN (or UCI) notation, and return a set of
    highlighted squares the move might be affecting.
    """
    # [QKNRB]?[a-h]?[1-8]?x?[a-h][1-8]=?[QNRB]
    def squareset(pc, f1, r1, x, f2, r2):
        hi1 = ~chess.SquareSet() if pc or f1 or r1 else chess.SquareSet()
        if pc is not None: hi1 &= board.pieces(chess.Piece.from_symbol(pc).piece_type, board.turn) 
        if x and not pc: hi1 &= board.pieces(chess.PAWN, board.turn)
        if f1 is not None: hi1 &= chess.BB_FILES[ord(f1) - ord("a")]
        if r1 is not None: hi1 &= chess.BB_RANKS[ord(r1) - ord("1")]
        hi2 = ~chess.SquareSet() if f2 or r2 else chess.SquareSet()
        if f2 is not None: hi2 &= chess.BB_FILES[ord(f2) - ord("a")]
        if r2 is not None: hi2 &= chess.BB_RANKS[ord(r2) - ord("1")]
        return hi1 | hi2
    match = rx_move.match(text)
    if not match: return chess.SquareSet()
    pc, f1, r1, x, f2, r2, pr = match.groups()
    ss = squareset(pc, f1, r1, x, f2, r2)
    if not x and not f2 and not r2: ss |= squareset(pc, None, None, None, f1, r1)
    return ss

piece_starting_count = [None, 8, 2, 2, 2, 1, 1]
piece_material = [None, 1, 3, 3, 5, 9, 100]

def ChessBoard(self, board, hi_squares, mv_squares, flip=False, evalbar=None, pv=()):
    white = ""
    black = ""
    material = 0
    marks = {}
    for i, uci in reversed(list(enumerate(pv[:3]))):
        marks[ord(uci[0]) - ord("a"), ord(uci[1]) - ord("1")] = chr(ord("1") + i)
        marks[ord(uci[2]) - ord("a"), ord(uci[3]) - ord("1")] = chr(ord("1") + i)
    for p in reversed(chess.PIECE_TYPES):
        n = len(board.pieces(p, True)) - len(board.pieces(p, False))
        material += piece_material[p] * n
        for i in range(n): white += self.piece_symbols[p]
        for i in range(-n): black += self.piece_symbols[p]
    if material > 0: white = f"{white} +{material}"
    if material < 0: black = f"+{-material} {black}"
    with im.Table(2, (1,1), (1,1), 2):
        whitename = self.aispec[1]["name"] if self.aispec[1] else self.user_name
        blackname = self.aispec[0]["name"] if self.aispec[0] else self.user_name
        with im.Row():
            with im.Cell(): pass
            with im.Cell(): im.Text(whitename, attr=self.attr_title)
            with im.Cell(): im.Text(blackname, attr=self.attr_title, align=2)
            with im.Cell(): pass
    with im.Table(2, (1,1), (1,1), 2):
        with im.Row():
            with im.Cell(): pass
            with im.Cell(): im.Text(white, attr=self.attr_subtitle)
            with im.Cell(): im.Text(black, attr=self.attr_subtitle, align=2)
            with im.Cell(): pass
    if evalbar:
        win, loss = evalbar
        bar_white = min(max(int(8*3*win), 1), 8*3 - 1)
        bar_black = 8*3 - min(max(int(8*3*loss), 1), 8*3 - 1)
    with im.Table(2,6,6,6,6,6,6,6,6,2):
        # Top frame
        with im.Row():
            with im.Cell(): im.Text("  ", attr=self.attr_border)
            for file in range(7, -1, -1) if flip else range(8):
                with im.Cell():
                    name = ord("a") + file
                    im.Text(f"   {name:c}  " if flip else "      ", attr=self.attr_border)
            with im.Cell(): im.Text("  ", attr=self.attr_border)
        # The board
        for rank in range(8) if flip else range(7, -1, -1):
            with im.Row():
                # Left frame
                with im.Cell():
                    if flip:
                        if evalbar:
                            aw = 3*rank + 0 < bar_white
                            al = 3*rank + 0 >= bar_black
                            bw = 3*rank + 1 < bar_white
                            bl = 3*rank + 1 >= bar_black
                            cw = 3*rank + 2 < bar_white
                            cl = 3*rank + 2 >= bar_black
                            a = self.attr_eval_w if aw else self.attr_eval_l if al else self.attr_eval_w
                            b = self.attr_eval_w if bw else self.attr_eval_l if bl else self.attr_eval_w
                            c = self.attr_eval_w if cw else self.attr_eval_l if cl else self.attr_eval_w
                            im.Text("█ " if aw or al else "░ ", attr=a)
                            im.Text("█ " if bw or bl else "░ ", attr=b)
                            im.Text("█ " if cw or cl else "░ ", attr=c)
                    else:
                        name = ord("1") + rank
                        im.Text("  ", attr=self.attr_border)
                        im.Text(f" {name:c}", attr=self.attr_border)
                        im.Text("  ", attr=self.attr_border)
                # Pieces
                for file in range(7, -1, -1) if flip else range(8):
                    white = (rank ^ file) & 1
                    with im.Cell():
                        if self.ai[self.board.turn] is None and \
                                not self.board.is_game_over(claim_draw=self.draw) and \
                                im.MouseClick(im.curx, im.cury, 6, 3):
                            square = chr(ord("a") + file) + chr(ord("1") + rank)
                            self.move = \
                                    self.move[:-2] if self.move.endswith(square) else \
                                    self.move + square + "\n"
                            self.try_user_move()
                            im.want_refresh = True
                        hi = chess.square(file, rank) in hi_squares
                        mv = chess.square(file, rank) in mv_squares
                        p = self.board.piece_at(chess.square(file, rank))
                        attr = (self.attr_hi_piece if hi else \
                                self.attr_mv_piece if mv else \
                                self.attr_piece)[white + 2*(p.color if p else white)]
                        for i, line in enumerate(self.piece_art[p.piece_type if p else 0]):
                            if i == 2 and (file,rank) in marks:
                                im.Text(marks[file,rank] + line[1:], attr=attr)
                            else:
                                im.Text(line, attr=attr)
                # Right frame
                with im.Cell():
                    if flip:
                        name = ord("1") + rank
                        im.Text("  ", attr=self.attr_border)
                        im.Text(f"{name:c} ", attr=self.attr_border)
                        im.Text("  ", attr=self.attr_border)
                    else:
                        if evalbar:
                            aw = 3*rank + 0 < bar_white
                            al = 3*rank + 0 >= bar_black
                            bw = 3*rank + 1 < bar_white
                            bl = 3*rank + 1 >= bar_black
                            cw = 3*rank + 2 < bar_white
                            cl = 3*rank + 2 >= bar_black
                            a = self.attr_eval_w if aw else self.attr_eval_l if al else self.attr_eval_w
                            b = self.attr_eval_w if bw else self.attr_eval_l if bl else self.attr_eval_w
                            c = self.attr_eval_w if cw else self.attr_eval_l if cl else self.attr_eval_w
                            im.Text(" █" if cw or cl else " ░", attr=c)
                            im.Text(" █" if bw or bl else " ░", attr=b)
                            im.Text(" █" if aw or al else " ░", attr=a)
        # Bottom frame
        with im.Row():
            with im.Cell(): im.Text("  ", attr=self.attr_border)
            for file in range(7, -1, -1) if flip else range(8):
                with im.Cell():
                    name = ord("a") + file
                    im.Text("      " if flip else f"  {name:c}   ", attr=self.attr_border)
            with im.Cell(): im.Text("  ", attr=self.attr_border)

class UI0:
    def __init__(self, config):
        self.config = config
        art = config["piece_art"]
        self.piece_art = [[art[0][i:i+6], art[1][i:i+6], art[2][i:i+6]] for i in range(0,8*6,6)]
        self.attr_piece = [
            # piece w/b -- square l/d
            config["style"]["square_wl"],  # 0b00
            config["style"]["square_wd"],  # 0b01
            config["style"]["square_bl"],  # 0b10
            config["style"]["square_bd"]   # 0b11
        ]
        self.attr_hi_piece = [
            config["style"]["square_hi_wl"], # 0b00
            config["style"]["square_hi_wd"], # 0b01
            config["style"]["square_hi_bl"], # 0b10
            config["style"]["square_hi_bd"]  # 0b11
        ]
        self.title = "??"
        self.ai_index = imui.Ref(0)
        self.color_index = imui.Ref(1)
        self.ailist = list(config["ai"].values())

    def render(self):
        with im.Center(width=60, height=10 + len(self.ailist) + len(self.config["rating_classes"])):
            im.Text("┳┓ ╭─┐┐           ")
            im.Text("┃┃ │  │  ╭─╮╭─┐╭─┐")
            im.Text("┣┻┓│  ├─╮├─╯╰─╮╰─╮")
            im.Text("┻━┛╰─╴┴ ┴╰─╴└─╯└─╯")
            im.VSpace(1)
            allais = []
            with im.TableRow(40, 14):
                with im.Cell():
                    im.Text("Play against:", attr=curses.A_BOLD)
                    with im.ListView(index=self.ai_index, id="lv-ai"):
                        for name, rat in self.config["rating_classes"].items():
                            ailist = [ai for ai in self.ailist if rat["min"] <= ai["rating"] < rat["max"]]
                            ailist.sort(key=lambda ai: ai["rating"])
                            allais.extend(ailist)
                            im.Text(f"== {name} ==")
                            #for i, ai in enumerate(self.ailist):
                            for i, ai in enumerate(ailist):
                                letter = chr(ord("1") + i if i < 9 else ord("a") + i - 9)
                                #if im.ListItem(f"{letter}) {ai['name']}", key=letter):
                                if im.ListItem(f"{letter}) {ai['name']}"):
                                    self.title = ai
                with im.Cell():
                    im.Text("Play as:", attr=curses.A_BOLD)
                    with im.TableRow(7,(1,1)):
                        with im.Cell():
                            attr = self.attr_piece[3 if self.color_index() == 2 else 1]
                            if im.Button("\n".join(self.piece_art[6 if self.color_index() != 1 else 7]), attr=attr):
                                self.color_index((self.color_index() + 1) % 3)
                                im.want_refresh = True
                        with im.Cell():
                            with im.ListView(index=self.color_index, id="lv-col"):
                                im.ListItem(f"Black", key="B")
                                im.ListItem(f"Random", key="R")
                                im.ListItem(f"White", key="W")
                    im.VSpace(1)
                    if im.Button("Start", key="\n", id="b-go", attr=curses.A_BOLD) or im.Key("\n"):
                        usercolor = \
                            False if self.color_index() == 0 else \
                            True if self.color_index() == 2 else \
                            random.choice((True,False))
                        if usercolor:
                            set_scene(UI(None, allais[self.ai_index()], self.config).render)
                        else:
                            set_scene(UI(allais[self.ai_index()], None, self.config).render)
            im.VSpace(1)
            aispec = allais[self.ai_index()]
            im.Text(f"{aispec['name']}", attr=curses.A_BOLD)
            FormatText(aispec.get("description", ""))

class UI:
    def __init__(self, white_ai, black_ai, config):
        self.attr_piece = [
            # piece w/b -- square l/d
            config["style"]["square_wl"], # 0b00
            config["style"]["square_wd"], # 0b01
            config["style"]["square_bl"], # 0b10
            config["style"]["square_bd"]  # 0b11
        ]
        self.attr_hi_piece = [
            config["style"]["square_hi_wl"], # 0b00
            config["style"]["square_hi_wd"], # 0b01
            config["style"]["square_hi_bl"], # 0b10
            config["style"]["square_hi_bd"]  # 0b11
        ]
        self.attr_mv_piece = [
            config["style"]["square_mv_wl"], # 0b00
            config["style"]["square_mv_wd"], # 0b01
            config["style"]["square_mv_bl"], # 0b10
            config["style"]["square_mv_bd"]  # 0b11
        ]
        self.attr_title = config["style"]["title"]
        self.attr_subtitle = config["style"]["subtitle"]
        self.attr_border = config["style"]["border"]
        self.attr_input = config["style"]["input"]
        self.attr_eval_w = config["style"]["eval_win"]
        self.attr_eval_d = config["style"]["eval_draw"]
        self.attr_eval_l = config["style"]["eval_loss"]
        art = config["piece_art"]
        self.piece_art = [[art[0][i:i+6], art[1][i:i+6], art[2][i:i+6]] for i in range(0,7*6,6)]
        self.piece_symbols = [None] + config["piece_symbols"]
        self.flip = False if white_ai is None else True
        self.help = False
        self.pgn_filename = config["pgn_filename"]
        self.user_name = os.environ.get("USER", "user")
        self.board = chess.Board()
        self.eval = {}
        self.aispec = (black_ai, white_ai)
        self.book = (
            book.of_spec(black_ai.get("book", None)) if black_ai else None,
            book.of_spec(white_ai.get("book", None)) if white_ai else None
        )
        self.ai = (engine.of_spec(black_ai) if black_ai else None, engine.of_spec(white_ai) if white_ai else None)
        self.eval_ai = engine.of_spec(config["eval_ai"])
        self.move = ""
        self.draw = False
        self.pgn_root = chess.pgn.Game()
        self.pgn = self.pgn_root
        self.prepare_ai_move()

    def save_pgn(self, filename):
        game = chess.pgn.Game()
        game.headers["White"] = self.aispec[1]["name"] if self.aispec[1] else self.user_name
        game.headers["Black"] = self.aispec[0]["name"] if self.aispec[0] else self.user_name
        game.headers["Date"] = datetime.date.today().strftime("%Y.%m.%d")
        game.headers["Event"] = "??"
        game.headers["Site"] = "??"
        game.headers["Round"] = "1"
        game.headers["Result"] = self.board.result(claim_draw=self.draw)
        if self.eval_ai:
            game.headers["Annotator"] = self.eval_ai.id.get("name", self.eval_ai.exepath)
        node = game
        board = chess.Board()
        opening = ""
        for move in self.board.move_stack:
            node = node.add_main_variation(move)
            board.push(move)
            ev = self.eval.get(board.fen(), None)
            if ev:
                ev = max(ev.items(), key=lambda kv: kv[0])[1]
                node.comment = f"[%eval {engine.score_eval(ev.score, ev.depth)}]"
            opening = ecodb.db.get(board.epd(), opening)
        eco, opening, variation = opening.split(":", 2)
        if eco: game.headers["ECO"] = eco
        if opening: game.headers["Opening"] = opening
        if variation: game.headers["Variation"] = variation
        with open(filename, "w") as f:
            f.write(str(game))
            f.write("\n\n\n")

    def apply_move(self, move):
        if move == "draw":
            if not self.board.can_claim_draw():
                raise ValueError("No basis to claim draw now.")
            self.draw = True
        elif move == "undo":
            if len(self.board.move_stack) < 2:
                raise ValueError("Undo what?")
            self.board.pop()
            self.board.pop()
        elif move == "flip":
            self.flip = not self.flip
        elif move in ("quit", "exit", "resign"):
            exit(0)
        elif move in "help":
            self.help = not self.help
        else:
            try:
                self.board.push_uci(move)
            except:
                self.board.push_san(move)
        if self.board.move_stack:
            self.save_pgn("/tmp/bchess.pgn")
        if not self.board.is_game_over(claim_draw=self.draw):
            if self.eval_ai:
                self.eval_ai.analyze(self.board, self.eval_ai_update, self.board.fen())
            self.prepare_ai_move()
        im.want_refresh = True

    def prepare_ai_move(self):
        ai = self.ai[self.board.turn]
        if ai:
            book = self.book[self.board.turn]
            bookmove = book(self.board) if book else None
            if bookmove:
                self.ai_update(bookmove)
            else:
                ai.play(self.board, self.ai_update)

    def try_user_move(self):
        if "\n" not in self.move: return
        # Extra logic for multi-line pasted moves.
        thismove, nextmove = self.move.split("\n", 1)
        try:
            self.apply_move(thismove)
            self.move = nextmove
            im.want_refresh = True
        except ValueError as e:
            self.move = thismove
            pass

    def render(self):
        hi_squares = highlight_san_move(self.move, self.board)
        mv_squares = chess.SquareSet()
        if self.board.move_stack:
            m = self.board.move_stack[-1]
            mv_squares.add(m.from_square)
            mv_squares.add(m.to_square)
        if self.board.is_check():
            mv_squares.add(self.board.king(self.board.turn))
        with im.Center(width=52, height=26+3-1):
            ev = self.eval.get(self.board.fen(), None)
            if self.help and ev:
                evalbar = [engine.score_winpercent(e.score) for d, e in sorted(ev.items())]
                evalbar = [min(evalbar[-4:]), 1-max(evalbar[-4:])]
                ev = max(ev.items(), key=lambda kv: kv[0])[1]
                ChessBoard(self, self.board, hi_squares, mv_squares, evalbar=evalbar, pv=ev.pv, flip=self.flip)
            else:
                ChessBoard(self, self.board, hi_squares, mv_squares, flip=self.flip)
            if self.board.is_game_over(claim_draw=self.draw):
                im.Text(self.board.result(claim_draw=self.draw), attr=self.attr_input, align=1)
            else:
                # Move input field
                if self.ai[self.board.turn] is None:
                    if self.board.can_claim_fifty_moves():
                        im.Text("You can now claim draw by the fifty-move rule.", align=1)
                    if self.board.can_claim_threefold_repetition():
                        im.Text("You can now claim draw by threefold repetition.", align=1)
                    im.VSpace(1)
                    prefix = f"Move {self.board.fullmove_number}. " if self.board.turn else \
                             f"Move {self.board.fullmove_number}... "
                    self.move, chg = im.Input(self.move, prefix=prefix, attr=self.attr_input, align=1)
                    if chg: im.want_refresh = True
                    if "\n" in self.move:
                        self.try_user_move()
                else:
                    im.VSpace(1)
                    self.move, chg = im.Input(self.move, prefix="Thinking... ", attr=self.attr_input, align=1)

    def ai_update(self, move):
        # The update here must be delayed because:
        # 1) ai_update might get called from an AI thread;
        # 2) ai_update might get called from apply_move itself.
        im.run_soon(lambda: self.apply_move(move))

    def eval_ai_update(self, result, fen):
        if isinstance(result, engine.Evaluation):
            if result.depth >= 4:
                self.eval.setdefault(fen, {})
                self.eval[fen][result.depth] = result
                im.want_refresh = True

### MAIN

def config_implement_colors(conf):
    colors = {(-1, -1): 0}
    colorn = 1
    for a in conf.values():
        fg = a.get("fg", -1)
        bg = a.get("bg", -1)
        if (fg, bg) in colors: continue
        curses.init_pair(colorn, fg, bg)
        colors[fg, bg] = curses.color_pair(colorn)
        colorn += 1
    result = {}
    for key, a in list(conf.items()):
        attr = colors[a.get("fg", -1), a.get("bg", -1)]
        attrstr = a.get("attr", "")
        if "a" in attrstr: attr |= curses.A_ALTCHARSET
        if "B" in attrstr: attr |= curses.A_BLINK
        if "b" in attrstr: attr |= curses.A_BOLD
        if "d" in attrstr: attr |= curses.A_DIM
        if "I" in attrstr: attr |= curses.A_INVIS
        if "r" in attrstr: attr |= curses.A_REVERSE
        if "s" in attrstr: attr |= curses.A_STANDOUT
        if "u" in attrstr: attr |= curses.A_UNDERLINE
        result[key] = attr
    return result

def curses_main(win):
    config = default_config()
    curses.curs_set(False)
    curses.start_color()
    curses.use_default_colors()
    config["style"] = config_implement_colors(config["style"])
    set_scene(UI0(config).render)
    curses.cbreak()
    curses.mousemask(-1)
    curses.mouseinterval(0)
    while True:
        with im.Frame(win):
            UI_RENDERER()
            if im.Key("q"):
                break
        while not (im.want_refresh or im.gather_input(win)):
            pass

def main():
    global im
    im = imui.IM()
    book.default = book.BookDB(config_subs("{data}/openings.sqlite"))
    try:
        curses.wrapper(curses_main)
    except KeyboardInterrupt:
        pass

if __name__ == "__main__":
    main()



import random
import sqlite3

class BookDB:
    def __init__(self, filename):
        self.db = None
        self.db = sqlite3.connect(f"file:{filename}?immutable=1", uri=True, check_same_thread=False)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def __del__(self):
        self.close()

    def close(self):
        if self.db:
            self.db.close()

    def available_ratings(self):
        return [elo for elo, in self.db.execute("select distinct elo from moves")]

    def moves(self, epd, elo):
        movestring = None
        for row in self.db.execute("select moves from moves where elo=? and boardid=(select id from boards where epd=?)", (elo, epd)):
            movestring, = row
        if movestring is None:
            return [], []
        moves = []
        counts = []
        for mc in movestring.split():
            move, count = mc.split(":")
            moves.append(move)
            counts.append(int(count))
        return moves, counts

    def random_move(self, epd, elo):
        moves, counts = self.moves(epd, elo)
        if moves == []:
            return None
        return random.choices(moves, weights=counts)[0]

    def evaluations(self, epd):
        for depth, score, pv in self.db.execute("select depth,score,pv from evaluations where boardid=(select id from boards where epd=?)", (epd,)):
            yield depth, score, pv

default = None

def of_spec(spec):
    if spec is None: return None
    assert spec["type"] == "builtin"
    assert default is not None
    rating = spec["rating"]
    return lambda board: default.random_move(board.epd(), rating)

def evaldb():
    assert default is not None
    return lambda board: list(default.evaluations(board.epd()))



import curses

from contextlib import contextmanager
from collections import namedtuple

EventMousePress = namedtuple("EventMousePress", "button x y")

class Ref:
    """
    Mutable cell with one value.

    One can get the value by calling this object with no parameters,
    and set it by calling it with one.

    See also: class RefOfAttribute and RefOfItem.
    """
    __slots__ = ("value",)
    def __init__(self, value):
        self.value = value
    def __call__(self, value=None):
        if value is not None: self.value = value
        return self.value

class RefOfAttribute:
    """
    Mutable cell referencing an attribute of an object.

    See also: class Ref and RefOfItem.
    """
    __slots__ = ("container", "key")
    def __init__(self, obj, key):
        self.obj = obj
        self.key = key
    def __call__(self, value=None):
        if value is not None: setattr(self.obj, key, value)
        return getattr(self.obj, self.key)

class RefOfItem:
    """
    Mutable cell referencing a keyed item of a container.

    See also: class Ref and RefOfAttribute.
    """
    __slots__ = ("container", "key")
    def __init__(self, container, key):
        self.container = container
        self.key = key
    def __call__(self, value=None):
        if value is not None: self.container[key] = value
        return self.container[self.key]

class IM_EndOfScreen(Exception):
    """
    Throw this exception when the bottom of the screen (or the
    current column) is reached, and further rendering will only
    go below it, so it can be safely skipped.
    """
    pass

class IM:
    """
    This is an "immediate mode" terminal UI library.

    IM is a bit specialized and not 100% general, but the approach
    adopted here is the best way to construct UIs, terminal or
    not.
    """

    def __init__(self):
        self.win = None
        self.key_events = []
        self.mouse_events = []
        self.want_refresh = False
        self.minx = 0
        self.miny = 0
        self.maxx = 0
        self.maxy = 0
        self.curx = 0
        self.cury = 0
        self.active_id = None
        self.active_index = None
        self.current_index = 0
        self.runqueue = []
        self.table_rowy, self.table_maxy, self.table_col_minx, self.table_col_maxx, self.table_idx = None, None, None, None, None

    def gather_input(self, win):
        """
        Read and process a part of the available input from stdin, 
        possibly pausing a few ms to wait for additional input.
        Execute all functions enqueued by run_soon() at the end.
        You should call this in a loop.
        """
        curses.halfdelay(7)
        win.nodelay(1)
        while True:
            try:
                key = win.get_wch()
            except curses.error:
                break
            curses.cbreak()
            if key == curses.KEY_MOUSE:
                try:
                    self.input_mouse(*curses.getmouse())
                except curses.error:
                    pass
            else:
                self.input_key(key)
        if self.runqueue:
            rq = self.runqueue
            self.runqueue = []
            for f in rq: f()
        return self.key_events != [] or self.mouse_events != []

    def run_soon(self, f):
        """Enqueue a function to be run after the next gather_input()."""
        self.runqueue.append(f)

    def input_mouse(self, mid, x, y, z, state):
        """Add a mouse event to the input queue."""
        if state & curses.BUTTON1_PRESSED:
            self.mouse_events.append(EventMousePress(1, x, y))
        if state & curses.BUTTON2_PRESSED:
            self.mouse_events.append(EventMousePress(2, x, y))
        if state & curses.BUTTON3_PRESSED:
            self.mouse_events.append(EventMousePress(3, x, y))
        if state & curses.BUTTON4_PRESSED:
            self.mouse_events.append(EventMousePress(4, x, y))
        if state & getattr(curses, "BUTTON5_PRESSED", 134217728):
            self.mouse_events.append(EventMousePress(5, x, y))

    def input_key(self, key):
        """Add a keyboard event to the input queue."""
        self.key_events.append(key)

    @property
    def text_width(self):
        """The width of the current column (or the screen)."""
        return self.maxx + 1 - self.minx

    @property
    def text_height(self):
        """The height of the current column (or the screen)."""
        return self.maxy + 1 - self.miny

    @contextmanager
    def Frame(self, screen):
        """
        Reset the screen state and prepare for a new frame. Call
        this each frame.
        """
        self.win = screen
        self.win.erase()
        self.want_refresh = False
        self._focus_begin()
        try:
            with self.Screen(): yield
        finally:
            self.win.refresh()
            self.win = None
            self.key_events = []
            self.mouse_events = []
            self._focus_end()

    @contextmanager
    def Screen(self):
        """
        Begin the layout of the whole screen (from the top left).
        Call only if an overlay over the previous layout needed,
        otherwise Frame() will call this automatically.
        """
        save = self.minx, self.miny, self.maxy, self.maxx, self.curx, self.cury
        self.minx, self.miny = 0, 0
        self.maxy, self.maxx = self.win.getmaxyx()
        self.maxy -= 1
        self.maxx -= 1
        self.curx, self.cury = 0, 0
        try:
            yield
        except IM_EndOfScreen:
            pass
        finally:
            self.minx, self.miny, self.maxy, self.maxx, self.curx, self.cury = save

    def Key(self, key, id=None):
        """
        An invisible key accelerator. Returns the number of times
        the key was pressed. If id is given, only consumes keys
        when the specified element is active.
        """
        if id is None or self.active_id == id:
            n = len(self.key_events)
            self.key_events = [e for e in self.key_events if e != key]
            return n - len(self.key_events)
        else:
            return 0

    def _focus_begin(self):
        """Call at the start of Frame() to update focus data."""
        if self.active_id is None:
            # Nothing is active yet, searching for active_index.
            n = self.Key("\t") - self.Key(curses.KEY_BTAB)
            if n:
                if self.active_index is None:
                    self.active_index = n - 1 if n > 0 else n
                else:
                    self.active_index += n
        if self.active_index is not None:
            if self.current_index > 0:
                self.active_index %= self.current_index
            else:
                self.active_index = None
        self.current_index = 0
        self.active_exists = False

    def _focus_end(self):
        """Call at the end of Frame() to update focus data."""
        if self.active_index is not None:
            # This might be off if the new active element was
            # not found during the frame. If so, try again.
            self.want_refresh = True
        if self.active_id is not None and not self.active_exists:
            self.active_id = None

    def cycle_focus(self, id):
        """
        Detect if Tab was pressed, and if so, change the focus
        to the next element. Every focusable element should call
        this exactly once, before determining focus.
        Return True if the current element is focused.
        """
        assert self.active_id is None or self.active_index is None
        if id is None:
            # This element is not even focusable.
            return False
        if self.active_id == id:
            # This element is active, unless Tab was pressed...
            n = self.Key("\t") - self.Key(curses.KEY_BTAB)
            if n:
                # ... in which case unfocus and search for the
                # next focused element.
                self.active_id = None
                self.active_index = self.current_index + n
            if self.Key("\x1b") != 0:
                # Or maybe just unfocus all.
                self.active_id = None
                self.active_index = None
        elif self.active_id is None:
            # No element is active.
            if self.active_index == self.current_index:
                self.active_id = id
                self.active_index = None
        else:
            # Some other element is active.
            pass
        self.current_index += 1
        if self.active_id == id:
            self.active_exists = True
            return True
        else:
            return False

    def focus(self, id):
        if id is None:
            return False
        if self.active_id is not None:
            self.want_refresh = True
        self.active_id = id
        self.active_index = None
        return True

    def AllKeys(self, id=None):
        if id is None or self.active_id == id:
            for k in self.key_events: yield k
            self.key_events = []
        else:
            return 0

    def Input(self, text, id=None, prefix="", attr=0, align=0):
        change = False
        if self.active_id == id:
            while self.key_events:
                key = self.key_events.pop(0)
                if key == curses.KEY_BACKSPACE:
                    if text == "":
                        pass
                    else:
                        text = text[:-1]
                        change = True
                elif type(key) == str:
                    text = text + key
                    change = True
            self.Text(prefix + text + "▒", attr=attr, align=align)
        else:
            self.Text(prefix + text + " ", attr=attr, align=align)
        return text, change

    def MouseClick(self, x, y, w, h):
        """
        An invisible mouse region. Returns the number of times
        it was clicked.
        """
        n = len(self.mouse_events)
        self.mouse_events = [
            e for e in self.mouse_events
            if not((x <= e.x < x + w) and (y <= e.y < y + h))
        ]
        return n - len(self.mouse_events)

    def TextAt(self, x, y, text, attr=0):
        """A text label at fixed coordinates."""
        w = len(text)
        if y <= self.maxy and x <= self.maxx:
            if x + w <= self.maxx:
                self.win.addstr(y, x, text, attr)
            else:
                try:
                    ll = self.maxx - x + 1
                    self.win.addstr(y, x, text[:ll], attr)
                except curses.error:
                    # Lower right corner always fails, no matter
                    # what. Stupid.
                    pass

    def Text(self, text, attr=0, align=0):
        """A text label (at the current layout position)."""
        text = str(text).strip("\"\'")
        if "\n" in text: _, text = text.rsplit("\n", 1)
        if align == 0:
            self.TextAt(self.curx, self.cury, text, attr=attr)
        elif align == 1:
            self.TextAt((self.curx + self.maxx - len(text) + 2)//2, self.cury, text, attr=attr)
        elif align == 2:
            self.TextAt(self.maxx - len(text) + 1, self.cury, text, attr=attr)
        self.cury += 1

    def VSpace(self, rows):
        """A number of empty rows."""
        self.cury += max(rows, 0)

    def HSpace(self, columns):
        """A number of empty columns"""
        self.curx += max(columns, 0)

    def HR(self, fill="-"):
        """A horizontal ruler, filling the whole line."""
        return self.Text(fill * (self.maxx - self.minx + 1))

    def Button(self, text, attr=0, hiattr=None, id=None, key=None):
        """
        A clickable and maybe pressable label. Returns the number
        of times it was clicked or pressed.
        """
        focus = self.cycle_focus(id)
        lines = text.splitlines()
        x, y, w, h = self.curx, self.cury, max(len(line) for line in lines), len(lines)
        hit = self.Key(key, id=id) + self.MouseClick(x, y, w, h)
        if hit: focus = self.focus(id)
        for line in lines:
            self.Text(line, (hiattr if hiattr is not None else attr | curses.A_REVERSE) if focus else attr)
        return hit

    @contextmanager
    def Table(self, *widths, margin=0):
        save = self.table_rowy, self.table_maxy, self.table_col_minx, self.table_col_maxx, self.table_idx
        minx, maxx = self.minx, self.maxx
        colw = [w if type(w) is int else w[0] for w in widths]
        colf = [0 if type(w) is int else w[1] for w in widths]
        self.table_rowy = self.cury
        self.table_maxy = self.cury
        self.table_col_minx = []
        self.table_col_maxx = []
        extraw = 1 + maxx - minx - sum(colw) - margin*(len(colw)-1)
        x = minx
        for i in range(len(colw)):
            flex = 0 if colf[i] == 0 else (extraw * colf[i] // sum(colf))
            self.table_col_minx.append(x)
            self.table_col_maxx.append(min(x + colw[i] + flex - 1, maxx))
            x += colw[i] + flex + margin
        yield
        self.minx, self.maxx = minx, maxx
        self.curx = minx
        self.cury = self.table_maxy
        self.table_rowy, self.table_maxy, self.table_col_minx, self.table_col_maxx, self.table_idx = save

    @contextmanager
    def Row(self):
        if self.table_rowy > self.maxy: raise IM_EndOfScreen()
        self.table_idx = 0
        self.table_maxy = 0
        yield True
        self.table_rowy = self.table_maxy

    @contextmanager
    def Cell(self, colspan=1):
        self.minx = self.table_col_minx[self.table_idx]
        self.maxx = self.table_col_maxx[self.table_idx + colspan - 1]
        self.curx = self.minx
        self.cury = self.table_rowy
        yield
        self.table_maxy = max(self.table_maxy, self.cury)
        self.table_idx += colspan

    def RowSeparator(self, fill="-"):
        with self.Row():
            for i in range(len(self.table_col_minx)):
                with self.Cell():
                    self.HR(fill)

    @contextmanager
    def TableRow(self, *widths, margin=0):
        with self.Table(*widths, margin=margin):
            with self.Row():
                yield

    @contextmanager
    def Center(self, width=None, height=None):
        if width is not None:
            with self.Table((0,1), width, (0,1), margin=0):
                with self.Row():
                    with self.Cell(): pass
                    with self.Cell():
                        if height is not None:
                            self.VSpace((self.text_height - height)//2)
                        yield
                    #with self.Cell(): pass
        else:
            if height is not None:
                self.VSpace((self.text_height - height)//2)
            yield

    @contextmanager
    def TabBar(self, *widths, index=None):
        self.tab_count = len(widths)
        self.tab_active = index() if index is not None else 0
        self.tab_current = 0
        hit = self.Key("\t") + self.Key(curses.KEY_RIGHT) - self.Key(curses.KEY_BTAB) - self.Key(curses.KEY_LEFT)
        self.tab_active = (self.tab_active + hit) % self.tab_count
        with self.Table(*widths):
            with self.Row():
                yield
        if index is not None:
            index(self.tab_active)

    @contextmanager
    def Tab(self, text, attr=0, key=None):
        with self.Cell():
            a = curses.A_REVERSE if self.tab_active == self.tab_current else attr
            if self.Button(text, attr=a, key=key):
                self.tab_active = self.tab_current
            self.tab_current += 1

    @contextmanager
    def ListView(self, index, id=None):
        self.lv_id = id
        self.lv_active = index() if index is not None else 0
        self.lv_current = 0
        if self.cycle_focus(id):
            hit = self.Key(curses.KEY_DOWN) - self.Key(curses.KEY_UP)
            self.lv_active += hit
        yield
        if not (0 <= self.lv_active < self.lv_current):
            self.lv_active %= self.lv_current
            self.want_refresh = True
        if index() != self.lv_active:
            index(self.lv_active)
            # This is only needed if some UI before the list
            # view depended on the index value.
            self.want_refresh = True

    def ListItem(self, text, attr=0, key=None):
        if self.Key(key):
            self.lv_active = self.lv_current
        if self.lv_active == self.lv_current:
            attr |= curses.A_REVERSE if self.active_id == self.lv_id else curses.A_UNDERLINE
        if self.Button(text, attr=attr):
            self.active_id = self.lv_id
            self.lv_active = self.lv_current
            self.want_refresh = True
        self.lv_current += 1

